// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sim.h"
#include <math.h>

void free_bands(struct device *in,double ** var)
{
	int i;
	for (i=0;i<in->ymeshpoints;i++)
	{
		free(var[i]);
	}

free(var);
}

void malloc_bands(struct device *in,double *(** var))
{
	*var = (double **)malloc(in->ymeshpoints*sizeof(double*));

	int i;
	for (i=0;i<in->ymeshpoints;i++)
	{
		(*var)[i] = (double *)malloc(in->bands*sizeof(double));
	}
}


void device_free(struct device *in)
{
	free(in->phi);
	free(in->Nad);
	free_bands(in,in->n);
	free_bands(in,in->p);
	free(in->ni);
	free_bands(in,in->dn);
	free_bands(in,in->dp);
	free(in->Eg);
	free(in->Xi);
	free(in->Ev);
	free(in->Ec);
	free_bands(in,in->mun);
	free_bands(in,in->mup);
	free(in->ymesh);
	free_bands(in,in->Fn);
	free_bands(in,in->Fp);
	free_bands(in,in->dndTe);

	free(in->Nc);
	free(in->Nv);
	free(in->me);
	free(in->mh);
	free(in->G);
	free(in->Gex);
	free(in->Tl);
	free(in->Te);
	free(in->Th);
	free(in->R);
	free(in->ymatlayers);
	free(in->Fi);
	free_bands(in,in->Jn);
	free_bands(in,in->Jp);
	free_bands(in,in->x);
	free_bands(in,in->t);
	free_bands(in,in->xp);
	free_bands(in,in->tp);
	free(in->ex);
	free(in->Dex);
	free(in->Hex);
	free(in->epsilonr);
	free(in->back_x);
	free(in->back_xp);
	free(in->back_phi);
	free(in->back_ex);
	free(in->kf);
	free(in->kd);
	free(in->kr);
	free(in->a);
	free(in->Rbi);
	free(in->kl);
	free(in->ke);
	free(in->kh);
	free(in->Hl);
	free(in->He);
	free(in->Hh);
	free(in->Hasorb);
	free(in->nlast);
	free(in->plast);
	free(in->Egphoton);
	free(in->Xiphoton);
	free(in->Ecphoton);
	free(in->Evphoton);

	free(in->nt);
	free(in->dnt);
	free(in->Fnt);
	free(in->Nct);
	free(in->xt);
	free(in->tt);
	free(in->tauncap);
	free(in->taunescape);
	free(in->dostype);
	free(in->Rbi_k);

	free(in->pt);
	free(in->dpt);
	free(in->Fpt);
	free(in->Nvt);
	free(in->xpt);
	free(in->tpt);
	free(in->taupcap);
	free(in->taupescape);
	free(in->A);
	free(in->deltaE);

	free(in->mun_avg);
	free(in->mup_avg);

	free(in->ninit);
	free(in->pinit);

	//fclose(in->gnuplot);

	free(in->meshdata);
	int i;
	printf("mat number= %d\n",in->mat.number);
	for (i=0;i<in->mat.number;i++)
	{
		mat_free(&(in->mat.l[i]));
	}

	mat_free(&(in->lcontact));
	mat_free(&(in->rcontact));
	free(in->mat.l);
	free(in->capeff);

	printf("Solved %i Equations\n",in->odes);
}


void device_get_memory(struct device *in)
{
	in->odes=0;
	in->Egphoton = malloc(in->ymeshpoints*sizeof(double));
	in->Xiphoton = malloc(in->ymeshpoints*sizeof(double));
	in->Ecphoton = malloc(in->ymeshpoints*sizeof(double));
	in->Evphoton = malloc(in->ymeshpoints*sizeof(double));

	in->ninit = malloc(in->ymeshpoints*sizeof(double));
	in->pinit = malloc(in->ymeshpoints*sizeof(double));

	in->Hasorb = malloc(in->ymeshpoints*sizeof(double));
	in->phi = malloc(in->ymeshpoints*sizeof(double));
	in->Nad = malloc(in->ymeshpoints*sizeof(double));
	malloc_bands(in,&(in->n));
	malloc_bands(in,&(in->p));
	in->ni = malloc(in->ymeshpoints*sizeof(double));
	malloc_bands(in,&(in->dn));
	malloc_bands(in,&(in->dp));
	in->Eg = malloc(in->ymeshpoints*sizeof(double));
	malloc_bands(in,&(in->Fn));
	malloc_bands(in,&(in->Fp));

	in->Xi = malloc(in->ymeshpoints*sizeof(double));
	in->Ev = malloc(in->ymeshpoints*sizeof(double));
	in->Ec = malloc(in->ymeshpoints*sizeof(double));
	malloc_bands(in,&(in->mun));
	malloc_bands(in,&(in->mup));

	in->Nc = malloc(in->ymeshpoints*sizeof(double));
	in->Nv = malloc(in->ymeshpoints*sizeof(double));
	in->me = malloc(in->ymeshpoints*sizeof(double));
	in->mh = malloc(in->ymeshpoints*sizeof(double));
	in->G = malloc(in->ymeshpoints*sizeof(double));
	in->Gex = malloc(in->ymeshpoints*sizeof(double));
	in->Tl = malloc (in->ymeshpoints * sizeof(double));
	in->Te = malloc (in->ymeshpoints * sizeof(double));
	in->Th = malloc (in->ymeshpoints * sizeof(double));

	in->R = malloc (in->ymeshpoints * sizeof(double));
	in->Fi = malloc (in->ymeshpoints * sizeof(double));
	malloc_bands(in,&(in->Jn));
	malloc_bands(in,&(in->Jp));

	malloc_bands(in,&(in->x));
	malloc_bands(in,&(in->t));

	malloc_bands(in,&(in->xp));
	malloc_bands(in,&(in->tp));
	malloc_bands(in,&(in->dndTe));

	in->mun_avg = malloc (in->ymeshpoints * sizeof(double));
	in->mup_avg = malloc (in->ymeshpoints * sizeof(double));

	in->kf = malloc (in->ymeshpoints * sizeof(double));
	in->kd = malloc (in->ymeshpoints * sizeof(double));
	in->kr = malloc (in->ymeshpoints * sizeof(double));
	in->a = malloc (in->ymeshpoints * sizeof(double));
	in->Rbi = malloc (in->ymeshpoints * sizeof(double));

	in->back_x = malloc (in->ymeshpoints * sizeof(double));
	in->back_xp = malloc (in->ymeshpoints * sizeof(double));
	in->back_phi = malloc(in->ymeshpoints*sizeof(double));
	in->back_ex = malloc(in->ymeshpoints*sizeof(double));


	in->ex = malloc (in->ymeshpoints * sizeof(double));
	in->Dex = malloc (in->ymeshpoints * sizeof(double));
	in->Hex = malloc (in->ymeshpoints * sizeof(double));
	in->epsilonr = malloc (in->ymeshpoints * sizeof(double));
	in->kl = malloc(in->ymeshpoints*sizeof(double));
	in->ke = malloc(in->ymeshpoints*sizeof(double));
	in->kh = malloc(in->ymeshpoints*sizeof(double));
	in->Hl = malloc(in->ymeshpoints*sizeof(double));
	in->He = malloc(in->ymeshpoints*sizeof(double));
	in->Hh = malloc(in->ymeshpoints*sizeof(double));

	in->nlast = malloc(in->ymeshpoints*sizeof(double));
	in->plast = malloc(in->ymeshpoints*sizeof(double));

	in->nt = malloc(in->ymeshpoints*sizeof(double));
	in->dnt = malloc(in->ymeshpoints*sizeof(double));
	in->Fnt = malloc(in->ymeshpoints*sizeof(double));
	in->Nct = malloc(in->ymeshpoints*sizeof(double));
	in->xt = malloc(in->ymeshpoints*sizeof(double));
	in->tt = malloc(in->ymeshpoints*sizeof(double));
	in->tauncap = malloc(in->ymeshpoints*sizeof(double));
	in->taunescape = malloc(in->ymeshpoints*sizeof(double));
	in->dostype = malloc(in->ymeshpoints*sizeof(double));

	in->pt = malloc(in->ymeshpoints*sizeof(double));
	in->dpt = malloc(in->ymeshpoints*sizeof(double));
	in->Fpt = malloc(in->ymeshpoints*sizeof(double));
	in->Nvt = malloc(in->ymeshpoints*sizeof(double));
	in->xpt = malloc(in->ymeshpoints*sizeof(double));
	in->tpt = malloc(in->ymeshpoints*sizeof(double));
	in->taupcap = malloc(in->ymeshpoints*sizeof(double));
	in->taupescape = malloc(in->ymeshpoints*sizeof(double));

	in->A = malloc(in->ymeshpoints*sizeof(double));
	in->deltaE = malloc(in->ymeshpoints*sizeof(double));
	in->capeff = malloc(in->ymeshpoints*sizeof(double));
	in->Rbi_k = malloc(in->ymeshpoints*sizeof(double));

}
